% examplehawkes

density = 0.1; % density of network
np = 30; % number of processes
MUmag = 1e-3; % frequency of starting interactions
Amag = 0.85; % maximum eigenvalue of A

gamma = @(t) exp(-t); % CHANGING THIS FROM "@(t) exp(-t)" REQUIRES A DIFFERENT "CIF" BELOW
igamma = @(t) 1-exp(-t); % integral of gamma
gammalen = 10; % we allow the substitution gamma(t)=0 if t>gammalen

Asupp = squareform(rand(np*(np-1)/2,1)<density); % generate a symmetric A
MU = MUmag*sum(Asupp,2); % MU proportional to number of relationships
A = nanmin(Amag/max(eig(double(Asupp))),Amag) * Asupp; % limit eigenvalues of A to Amag, and values to no greater than Amag

nevt = ceil(100*density*np^2); % number of events to simulate
sreg = 0.1*nevt/np; % sparsity regularization coeffcient
prior = ~eye(np); % don't allow self-excitation

wrand = @(w) sum([0;cumsum(w(:))]<rand(1)*sum(w(:))); % weighted random integer generation
genMKF = @(m,a) @(t,ht,hm) wrand(m + a(:,hm)*gamma(t-ht));
solveICIF = @(m,a,targ) lambertwlog(log(a/m)+(a-targ)/m) - (a-targ)/m; % specifically works for gamma(t) = exp(-t);
genPPSIMFUN = @(summ,suma) @(tcurr,ht,hm,targ,~) tcurr + solveICIF(summ,suma(hm)*gamma(tcurr-ht),targ); % requires solveICIF
genCIF = @(summ,suma) @(tcurr,ht,hm) summ + suma(hm)*gamma(tcurr-ht); % standalone CIF


MUs = sum(MU);
As = sum(A,1);
CIF = genPPSIMFUN(MUs,As); % not actually a CIF, but a function to find the next event time %%%% REQUIRES "gamma = @(t) exp(-t)"
% CIF = genCIF(MUs,As); % actual CIF - markpointproc3() integrates this to generate events - works if gamma is a continous function
MKF = genMKF(MU,A);

[evt,evm] = markpointproc(CIF,MKF,nevt,1,gammalen); % generate event times (evt) and processes (evm)
endt = evt(end)*(1+1/nevt); % the end time of our simulation (a little after the last event) - time starts at 0
[Ar,MUr] = hawkes_recover({np,evt,evm},endt,gamma,igamma,sreg,prior,gammalen);

subplot(1,2,1), imagesc(A); title('model A')
subplot(1,2,2), imagesc(Ar); title('recovered A')