% relaynetworksim
clear

rng(10)

nnode = 20;
nevchain = 2000;
startrate = 0.1;
respondtime = [1e-10,2];

c = [0;1;zeros(nnode-3,1);1];
ring = toeplitz(c);
randlink = squareform(rand(nnode*(nnode-1)/2,1)<0.03);
interconnect = double(ring | randlink);

ndist = zeros(nnode,nnode);
currreach = logical(eye(nnode));
for i = 1:nnode
	nextreach = logical(interconnect*currreach) | currreach;
	ndist(nextreach & ~currreach) = i;
	if isequal(currreach,nextreach)
		break;
	end
	currreach = nextreach;
end
if ~all(currreach)
	error('graph appears to be disconnected')
end
% sourcesink = (accumarray([mod((0:nnode-1)' + randi(nnode-1,nnode,1),nnode)+1,(1:nnode)'],true,[nnode,nnode])... % have a sink for every source
% 	| rand(nnode)<0.05) & ~eye(nnode); % and some extra, but never oneself
sourcesink = rand(nnode)<0.05 & ~eye(nnode); % random sources and sinks

tic
evtstart = nevchain/startrate * sort(rand(nevchain,1));
[sinks,sources] = find(sourcesink);
i = randi(numel(sources),size(evtstart));
evpsource = sources(i);
evpsink = sinks(i);

evrec = cell(nevchain,1);
for i = 1:nevchain % generate events
	evtcurr = evtstart(i);
	evpcurr = evpsource(i);
	evtc = zeros(0,1);
	evpc = zeros(0,1);
	j = 0;
	while evpcurr ~= evpsink(i)
		j = j+1;
		evtc(j,1) = evtcurr;
		evpc(j,1) = evpcurr;
		candidates = find(interconnect(:,evpcurr) & ndist(:,evpsink(i)) < ndist(evpcurr,evpsink(i)));
		evpcurr = candidates(randi(numel(candidates))); % move to next node
		evtcurr = evtcurr + respondtime(1) + (respondtime(2)-respondtime(1))*rand; % response time of next node
	end
	j = j+1; evtc(j,1) = evtcurr; evpc(j,1) = evpcurr; % add sorta-acknowledgement from end node
	evrec{i} = [evtc,evpc,[evpc(2:end);0],repmat(i,j,1)];
end

[ev,ord] = sortrows(cell2mat(evrec));
evt = ev(:,1); evp = ev(:,2); evnext = ev(:,3); evtxind = ev(:,4);
tmax = evt(end) + startrate;
nevs = numel(evt);

cnevs = cellfun(@(x) size(x,1),evrec);
longestchain = max(cnevs);
cc = cell2mat(arrayfun(@colon,ones(nevchain,1),cnevs,'uni',false)')';
chainevs = accumarray([evtxind,cc(ord)],1:nevs,[nevchain,longestchain]); % get zero-padded event chains

evc = {nnode,evt,evp};
toc
nonterminal = find(evnext~=0);
At = accumarray([evnext(nonterminal),evp(nonterminal)],1,[nnode,nnode]); % ground truth for As, Ag
% Aet = accumarray([evnext(nonterminal),nonterminal],1,[nnode,nevs]); % ground truth for Aes
MUt = accumarray(evpsource,1,[nnode,1])/tmax; % ground truth for MU
%%

prior = ~eye(nnode);
if false
	tconst = respondtime(2)/2;
	gamma = @(t) exp(-t/tconst)/tconst;
	igamma = @(t) 1-exp(-t/tconst);
	gammalen = 10*tconst;
else
	gamma = @(t) (t>respondtime(1) & t<respondtime(2))/(respondtime(2)-respondtime(1));
	igamma = @(t) (t>respondtime(1) & t<respondtime(2)) .* (t-respondtime(1))/(respondtime(2)-respondtime(1)) + (t>respondtime(2));
	gammalen = respondtime(2);
end

regl1 = 0;

tic
[As,MUs] = hawkes_recover(evc,tmax,gamma,igamma,regl1,prior,gammalen);
toc

%%
tic

gmatijv = cell(nnode,1);
pbegin = zeros(nnode,1);
sgmatijv = cell(nnode,1);
spbegin = zeros(nnode,1);
gtgmatijv = cell(nnode,1);
gtpbegin = zeros(nnode,1);
for i = 1:nnode
	pinds = find(evp==i);
	[rsupp,csupp] = suppgmat(evt,pinds,gammalen);
	
	if isempty(rsupp)
		continue;
	end
	procnevs = numel(pinds);
	pcsupp = evp(csupp);
	prsupp = pinds(rsupp);
	val = gamma(evt(prsupp)-evt(csupp));
	
% 	gmati = sparse(rsupp,evp(csupp),val,procnevs,nnode); % small enough that we don't need sparse()
	gmati = accumarray([rsupp,pcsupp],val,[procnevs,nnode]);
	rtotal = MUs(i) + gmati*As(i,:)';
	gtrtotal = MUt(i) + gmati*At(i,:)';
	
	valevtevt = As(i,pcsupp)'.*val./rtotal(rsupp);
	pbegin(pinds) = MUs(i)./rtotal;
	gmatijv{i} = [prsupp,csupp,val,valevtevt];
	
	gtvalevtevt = At(i,pcsupp)'.*val./gtrtotal(rsupp);
	gtpbegin(pinds) = MUt(i)./gtrtotal;
	gtgmatijv{i} = [prsupp,csupp,val,gtvalevtevt];
end
gmatijv = cell2mat(gmatijv);
sgmatijv = cell2mat(sgmatijv);
gtgmatijv = cell2mat(gtgmatijv);
toc
tic
% gmat = sparse(gmatijv(:,1),gmatijv(:,2),gmatijv(:,3),nevs,nevs);
pend = exp(accumarray(gmatijv(:,2),log1p(-gmatijv(:,4)),[nevs,1]));
% pevtevt = sparse(gmatijv(:,1),gmatijv(:,2),gmatijv(:,4),nevs,nevs); % pevtevt(causee,causer)
gmatijvss = [gmatijv(:,[1,2,4]);[(1:nevs)',repmat(nevs+1,nevs,1),pbegin; repmat(nevs+2,nevs,1),(1:nevs)',pend]]; % augment with start and stop nodes
pevtevtss = sparse(gmatijvss(:,1),gmatijvss(:,2),gmatijvss(:,3),nevs+2,nevs+2);

% [paths,pathprobs] = highprobpaths(pevtevtss,1e-5,nevs+1,nevs+2,'firstlast');
% evstopstart = fliplr(paths);
% pthr = 0.5;
% nstartstop = accumarray(evp(evstopstart),pathprobs.*(pathprobs>=pthr),[nnode,nnode]);

[pathevs,pathprobs] = highprobpaths(pevtevtss,1e-4,nevs+1,nevs+2,'array');
evpstopstart = [evp(max(pathevs,[],2)),evp(pathevs(:,1))];
if size(pathevs,2) > longestchain
	shortenough = pathevs(:,longestchain+1) == 0;
else
	shortenough = true(numel(pathprobs),1);
end
ismatch = ismember(pathevs(shortenough,1:longestchain),chainevs,'rows');
truechainsfound = false(numel(pathprobs),1);
truechainsfound(shortenough) = ismatch;

pthr = 0.5;
nstartstop = accumarray(evpstopstart,pathprobs.*(pathprobs>=pthr),[nnode,nnode]);

toc

gtstartstop = accumarray([evpsink,evpsource],1,[nnode,nnode]);
%%

figure(1),clf(1),set(1,'position',[400,400,500,500])
ny = 2; nx = 2;
% subplot(ny,nx,1),imagesc(interconnect),xlabel('from'),ylabel('to'),title('base connectivity')
subplot(ny,nx,1),imagesc(At),xlabel('from'),ylabel('to'),title('True utilization')
subplot(ny,nx,2),imagesc(gtstartstop),xlabel('source'),ylabel('sink'),title('True sources/sinks')

subplot(ny,nx,3),imagesc(As),xlabel('from'),ylabel('to'),title('Recovered utilization')
subplot(ny,nx,4),imagesc(nstartstop),xlabel('source'),ylabel('sink'),title('Recovered sources/sinks')

sum(sum(gtstartstop.*nstartstop))./(norm(gtstartstop,'fro')*norm(nstartstop,'fro'))
sum(sum(gtstartstop.*nstartstop))./sum(sum(gtstartstop.*gtstartstop))
sum(pathprobs(truechainsfound))/nevchain % total correct probability found

binedges = (0:0.05:1)';
bincenters = conv(binedges,[0.5;0.5],'valid');
% probhistwrong = histcounts(pathprobs(~truechainsfound),binedges);
% probhistcorrect = histcounts(pathprobs(truechainsfound),binedges);
weightedhistwrong = accumarray(discretize(pathprobs(~truechainsfound),binedges),pathprobs(~truechainsfound),[numel(binedges)-1,1]);
weightedhistcorrect = accumarray(discretize(pathprobs(truechainsfound),binedges),pathprobs(truechainsfound),[numel(binedges)-1,1]);

figure(2),clf,set(gcf,'position',[400,400,300,300])

probthr = 0:1e-3:1;
thrpts = [1:2,4:3:16,18:19]'/20; % probabilities to mark on the ROC
fpw = cumsum(accumarray(discretize(pathprobs(~truechainsfound),probthr),pathprobs(~truechainsfound),[numel(probthr),1]),'reverse');
tpw = cumsum(accumarray(discretize(pathprobs(truechainsfound),probthr),pathprobs(truechainsfound),[numel(probthr),1]),'reverse');
fpwn = fpw/fpw(1);
tpwn = tpw/tpw(1);
plotpts = discretize(thrpts,probthr);
plot(fpwn,tpwn,[0;1],[0;1],'--',fpwn(plotpts),tpwn(plotpts),'k.'),xlabel('false positive weight'),ylabel('true positive weight'),title('Reciever Operating Characteristic'),grid on
text(fpwn(plotpts),tpwn(plotpts),num2str(thrpts),'VerticalAlignment','cap')
% axis equal